from .agent import SingleRoomAgent
from meshagent.api import websocket_protocol, RoomClient, MeshSchema, RoomException
import asyncio
import signal
import json

from aiohttp import web


async def connect_development_agent(*, room_name:str, agent: SingleRoomAgent):
    
    async with RoomClient(
        protocol = websocket_protocol(participant_name=agent.name, room_name=room_name, role="agent")) as room:
        
        await agent.start(room=room)

        try:
            
            term = asyncio.Future()

            def clean_termination(signal, frame):
                term.set_result(True)
            
            signal.signal(signal.SIGTERM, clean_termination)
            signal.signal(signal.SIGABRT, clean_termination)

            await asyncio.wait([asyncio.ensure_future(room.protocol.wait_for_close()), term], return_when=asyncio.FIRST_COMPLETED)

        finally:

            await agent.stop()




from abc import ABC, abstractmethod
from .agent import AgentChatContext
from jsonschema import validate
from meshagent.tools.toolkit import Response, Toolkit
from meshagent.api import RoomClient
from typing import Any, Optional, Callable, TypeVar, Generic

T = TypeVar("T")

class ToolResponseAdapter(ABC):
    def __init__(self):
        pass
    
    @abstractmethod
    async def to_plain_text(self, *, room: RoomClient, response: Response):
        pass

    @abstractmethod
    async def create_messages(self, *, context: AgentChatContext, tool_call: Any, room: RoomClient, response: Response) -> list:
        pass


class LLMAdapter(Generic[T]):

    def create_chat_context(self) -> AgentChatContext:
        return AgentChatContext()
    
    @abstractmethod
    async def check_for_termination(self, *, context: AgentChatContext, room: RoomClient):
        return True

    @abstractmethod
    async def next(self, 
        *,
        context: AgentChatContext,
        room: RoomClient,
        toolkits: Toolkit,
        tool_adapter: Optional[ToolResponseAdapter] = None,
        output_schema: Optional[dict] = None,
        event_handler: Optional[Callable[[T],None]] = None
    ) -> Any:  
        pass

    def validate(response: dict, output_schema: dict):
        validate(response, output_schema)




from .agent import SingleRoomAgent
from meshagent.api.chan import Chan
from meshagent.api import RoomMessage, RoomException, RoomClient, RemoteParticipant
from meshagent.agents import AgentChatContext
from meshagent.tools import Toolkit
from .adapter import LLMAdapter, ToolResponseAdapter
import asyncio
from typing import Optional
import json
from meshagent.tools import ToolContext
import logging

logger = logging.getLogger("chat")


class Worker(SingleRoomAgent):
    def __init__(self, *, queue: str,  name, title = None, description = None, requires = None, llm_adapter: LLMAdapter, tool_adapter:  Optional[ToolResponseAdapter] = None, toolkits: Optional[list[Toolkit]] = None, rules : Optional[list[str]] = None):
        super().__init__(
            name=name,
            title=title,
            description=description,
            requires=requires,
        )

        self._queue = queue

        if toolkits == None:
            toolkits = []

        self._llm_adapter = llm_adapter
        self._tool_adapter = tool_adapter

        self._message_channel = Chan[RoomMessage]()

        self._room : RoomClient | None = None
        self._toolkits = toolkits

        if rules == None:
            rules = []

        self._rules = rules
        self._done = False
       

    async def start(self, *, room: RoomClient):
        self._done = False

        await super().start(room=room)

        
        self._main_task = asyncio.create_task(self.run(room=room))

    async def stop(self):

        self._done = True

        await asyncio.gather(self._main_task)

        await super().stop()


    async def append_message_context(self, *, room: RoomClient, message: dict, chat_context: AgentChatContext):
        chat_context.append_user_message(message=json.dumps(message))
        
    async def process_message(self, *, chat_context: AgentChatContext, room: RoomClient, message: dict, toolkits: list[Toolkit]):

        return await self._llm_adapter.next(
            context=chat_context,
            room=room,
            toolkits=toolkits,
            tool_adapter=self._tool_adapter,
        )


    async def run(self, *, room: RoomClient):

        toolkits = [
            *await self.get_required_toolkits(ToolContext(room=room, caller=room.local_participant)),
            *self._toolkits
        ]
        
        while not self._done:

            message = await room.queues.receive(name=self._queue, create=True, wait=True)
            if message != None:
                
                logger.info(f"received message on worker queue {message}")
                try:

                    chat_context = await self.init_chat_context()

                    chat_context.append_rules(
                        rules=[
                            *self._rules,
                        ]
                    )
                    
                    await self.append_message_context(room=room, message=message, chat_context=chat_context)
                        

                    await self.process_message(chat_context=chat_context, room=room, message=message, toolkits=toolkits)
       
                except Exception as e:

                    logger.error(f"Failed to process a message {message}", exc_info=e)



from meshagent.agents.agent import AgentCallContext, AgentChatContext, AgentException
from meshagent.api import WebSocketClientProtocol, RequiredToolkit, RequiredSchema, Requirement
from meshagent.api.schema_document import Document
from meshagent.api.room_server_client import RoomClient
from meshagent.tools.toolkit import Toolkit, TextResponse, Tool, ToolContext
from meshagent.api.schema import MeshSchema
from meshagent.agents.writer import Writer, WriterContext
from meshagent.agents.adapter import LLMAdapter, ToolResponseAdapter
from meshagent.api.schema import MeshSchema, ElementType, ChildProperty, ValueProperty
from meshagent.api.schema_util import merge
from meshagent.tools.document_tools import build_tools, DocumentAuthoringToolkit
from meshagent.agents import TaskRunner
from copy import deepcopy
from typing import Optional

reasoning_rules = [
    "If an ask_user tool call is available, plans should include a series of questions to ask the user to help refine.",
    "If an ask_user tool call is available, ask a maximum of one question per step",
    "If an ask_user tool call is not available, you may not ask the user any questions",
    "You will be given a task. First formulate a plan for the task. Then execute the steps until you have a final answer.",
    "Do not use tool calls to write estimates of progress or plans",
    "You are a document generation service",
    "The user is asking for a document to be created",
    "You must use tool calls must to generate the answer to the user's question as document"
]

goto_next_step_message = """
     execute the next step, and provide the result of the step.
"""

def is_reasoning_done(*, context: AgentCallContext, response: dict) -> bool:
    parsed = response["response"]["data"][0]
    if "abort" in parsed:
        abort = parsed["abort"]
        reason = abort["reason"]
        raise AgentException(reason)
        
    elif "progress" in parsed:
        result = parsed["progress"]
                            
        if "done" not in result or result["done"]:
            logger.info("Done generating response %s", result)
            return True
        else:
            context.chat.append_user_message(goto_next_step_message)
            return False
        
    elif "plan" in parsed:
        plan = parsed["plan"]
        context.chat.append_user_message(goto_next_step_message)
        return False
    else:
        logger.info("recieved invalid response, %s", parsed)
        context.chat.append_user_message("this response did not conform to the schema")
        return False

def reasoning_schema(*, description: str, elements: Optional[list[ElementType]] = None, has_done_property: bool = True, has_abort: bool = True) -> MeshSchema:

    if elements == None:
        elements = []

    progress_properties = [
        ValueProperty(name="percentage", description="an estimate for how complete the task is so far", type="string"),
        ValueProperty(name="next", description="a very short description of the next step", type="string"),
    ]

    if has_done_property:
        progress_properties.append( ValueProperty(name="done", description="whether there is more work to do. the program will continue will continue to send messages to the LLM to refine the answer until done is set to true.", type="boolean") )

    elements = elements.copy()

    if has_abort:
        elements.append(ElementType(
                tag_name="abort",
                description="return if the task cannot completed because the user cancelled a request or errors could not be resolved",
                properties=[
                    ValueProperty(name="reason", description="the reason the task was aborted", type="string")
                ]
            )),

    return MeshSchema(
        root_tag_name="response",
        elements=[
            ElementType(
                tag_name="response",
                description="a response for a task",
                properties=[
                    ChildProperty(name="data", description="the response for a task, should contain a single item", child_tag_names=[
                        "plan","progress",
                        *map(lambda x: x.tag_name, elements)    
                    ])
                ]
            ),
            ElementType(
                tag_name="plan",
                description="a plan will be output for each task to describe the work that will be done, the work will be performed using tool calls.",
                properties=[
                    ChildProperty(name="steps", description="the steps for the plan", child_tag_names=["step"])
                ]
            ),
            ElementType(
                tag_name="step",
                description="a step in the plan",
                properties=[
                    ValueProperty(name="description", description="a short sentence description description of the work that will be performed to complete the user's request.", type="string")
                ],
            ),
            ElementType(
                tag_name="progress",
                description="the progress of the task",
                properties=progress_properties
            ),
            ElementType(
                tag_name="thinking",
                description="use to log information that will not be included in the final answer",
                properties=[
                    ValueProperty(name="text", description="used to log thoughts or progress", type="string"),
                ],
            ),
            *elements
        ])

from meshagent.api.schema_util import prompt_schema
import logging

logger = logging.getLogger("planning_agent")



class PlanningWriter(Writer):
    
    def __init__(self, *, name: str, llm_adapter: LLMAdapter, tool_adapter:  Optional[ToolResponseAdapter] = None, max_iterations : int = 100, toolkits: Optional[list[Tool]] = None, title: Optional[str] = None, description: Optional[str] = None, rules: Optional[list[str]] | None = None, requires: Optional[list[Requirement]] = None, supports_tools: Optional[bool] = None):
        
        super().__init__(
            name=name,
            description=description,
            title=title,
            input_schema=merge(
                schema=prompt_schema(description="use a prompt to generate content"), 
                additional_properties={
                "path" : { "type" : "string" }
            }),
            output_schema={
                "type" : "object",
                "additionalProperties" : False,
                "required" : [],
                "properties" : {
                }
            },
            requires = requires,
            supports_tools = supports_tools
        )

        if rules == None:
            rules = []

        self._rules = rules

        self._llm_adapter = llm_adapter
        self._tool_adapter = tool_adapter 
        self._max_iterations = max_iterations
        if toolkits == None:
            toolkits = []
        self.toolkits = toolkits

        self._planning_rules : list[str] = [
            *reasoning_rules,
            *rules
        ]  

    async def init_chat_context(self):
        chat = await super().init_chat_context()

        all_rules = self._planning_rules.copy()
        chat.append_rules(rules=all_rules)
        return chat
        
    async def write(self, writer_context: WriterContext, arguments: dict):

        writer_context.call_context.chat.append_rules(f"your are writing to the document at the path {writer_context.path}")

        arguments = arguments.copy()
        self.pop_path(arguments=arguments)
        
        execute = goto_next_step_message

        react = """
        based on what you know know, either execute the next task or formulate a new plan. If you have sufficient information to complete the task, return a final answer.
        """
    
        prompt = arguments["prompt"]

        writer_context.call_context.chat.append_user_message(message=prompt)

        rs = reasoning_schema(description="uses tools", elements=[]).to_json()

        i = 0
        while i < self._max_iterations:
            i += 1
            
            try:
                logger.info("COMPLETION STARTING: Step %s", i)
     
                base_args = arguments.copy()
                base_args.pop("path")
                
                toolkits = [
                    DocumentAuthoringToolkit(),
                    Toolkit(
                        name="meshagent.planning-writer.tools",
                        tools=build_tools(document_type="document", schema=writer_context.document.schema, documents={writer_context.path : writer_context.document})),
                    *self.toolkits,
                    *writer_context.call_context.toolkits
                ]

                responses = await self._llm_adapter.next(context=writer_context.call_context.chat, room=writer_context.room, toolkits=toolkits, tool_adapter=self._tool_adapter, output_schema=rs)

                logger.info("COMPLETION RESPONSE %s", responses)
            
            except Exception as e:
                logger.error("Unable to execute reasoning completion task", exc_info=e)
                # retry
                raise(e)
            
            parsed = responses["response"]["data"][0]
            if "abort" in parsed:
                abort = parsed["abort"]
                reason = abort["reason"]
                raise AgentException(reason)
                
            elif "progress" in parsed:
                result = parsed["progress"]

                                    
                if "done" not in result or result["done"]:
                    logger.info("Done generating response %s", result)
                    return {}
                else:
                    writer_context.call_context.chat.append_user_message(execute)
                    continue
            elif "plan" in parsed:
                plan = parsed["plan"]
                writer_context.call_context.chat.append_user_message(execute)
                continue
            else:
                logger.info("recieved invalid response, %s", parsed)
                writer_context.call_context.chat.append_user_message("this response did not conform to the schema")
                continue

 
class PlanningResponder(TaskRunner):

    def __init__(self, *, name:str, llm_adapter: LLMAdapter, tool_adapter:  Optional[ToolResponseAdapter] = None, output_schema: dict,  max_iterations : int = 100, toolkits: Optional[list[Toolkit]] = None, title: Optional[str] = None, description: Optional[str] = None, requires: Optional[list[Requirement]] = None, supports_tools : bool = True, input_prompt: bool = True, rules: Optional[list[str]] = None, labels: Optional[list[str]] = None):
        if isinstance(output_schema, dict) == False:
            raise Exception("schema must be a dict, got: {type}".format(type=type(output_schema)))

        self._input_prompt = input_prompt

        if rules == None:
            rules = []


        if input_prompt:
            input_schema = prompt_schema(description="use a prompt to generate content")
        else:
            input_schema = {
                "type" : "object",
                "additionalProperties" : False,
                "required" : [],
                "properties" : {
                }
            }

        super().__init__(
            name=name,
            title=title,
            description=description,
            input_schema=input_schema,
            output_schema=output_schema,
            requires=requires,
            supports_tools=supports_tools,
            labels=labels,
            toolkits=toolkits
        )
        
        self._max_iterations = max_iterations

        self._planning_rules : list[str] = [
            *rules,
            *reasoning_rules
        ]

        self._responses = dict()

        self._llm_adapter = llm_adapter
        self._tool_adapter = tool_adapter 

    async def init_chat_context(self):
        chat = self._llm_adapter.create_chat_context()

        all_rules = self._planning_rules.copy()
        chat.append_rules(rules=all_rules)
        return chat


    async def ask(self, context: AgentCallContext, arguments: dict):

        class ResponseTool(Tool):
            def __init__(self, output_schema: dict, context: AgentCallContext, parent:PlanningResponder):
                super().__init__(
                    name="respond",
                    title="respond", 
                    description="send the response to the user",     
                    input_schema=output_schema
                )
                self.parent = parent
                self.context = context 

            async def execute(self, *, context: ToolContext, **kwargs):
                self.parent._responses[self.context] = kwargs
                return TextResponse(text="the response was sent")
            
        class ResponseToolkit(Toolkit):
            def __init__(self, output_schema, context, parent):

                tools = [    
                    ResponseTool(output_schema=output_schema, context=context, parent=parent)
                ]

                super().__init__(
                    name="meshagent.responder",
                    title="responder",
                    description="tools for responding",
                    tools=tools
                )
                

        context.toolkits.append(ResponseToolkit(output_schema=self.output_schema, context=context, parent=self))
        
        execute = goto_next_step_message

        react = """
        based on what you know know, either execute the next task or formulate a new plan. If you have sufficient information to complete the task, return a final answer.
        """
    
        rs = reasoning_schema(description="uses tools", elements=[], has_done_property=True, has_abort=True).to_json()

        if self._input_prompt:
            prompt = arguments["prompt"]
            context.chat.append_user_message(message=prompt)

        room = context.room
        i = 0
        while i < self._max_iterations:
            i += 1
            
            try:
                logger.info("COMPLETION STARTING: Step %s", i)
                    
                responses = await self._llm_adapter.next(context=context.chat, room=room, toolkits=context.toolkits, tool_adapter=self._tool_adapter, output_schema=rs)

                logger.info("COMPLETION RESPONSE %s", responses)
            
            except Exception as e:
                logger.error("Unable to execute reasoning completion task", exc_info=e)
                # retry
                raise(e)
            
            parsed = responses["response"]["data"][0]

            if "abort" in parsed:
                abort = parsed["abort"]
                reason = abort["reason"]
                raise AgentException(reason)
                
            elif "progress" in parsed:
                result = parsed["progress"]

                if "done" in result and result["done"]:                   
                    if context not in self._responses:
                        context.chat.append_user_message("you must call the respond tool")
                        continue
                    
                    final_answer = self._responses.pop(context)
                    logger.info("Done generating response %s", final_answer)
                    return final_answer
                
                else:
                    context.chat.append_user_message(execute)
                    continue
            elif "plan" in parsed:
                plan = parsed["plan"]
                context.chat.append_user_message(execute)
                continue
            else:
                logger.info("recieved invalid response, %s", parsed)
                context.chat.append_user_message("this response did not conform to the schema")
                continue

        
        return {}

class DynamicPlanningResponder(TaskRunner):

    def __init__(self, *, name: str, llm_adapter: LLMAdapter, tool_adapter:  Optional[ToolResponseAdapter] = None,  max_iterations : int = 100, toolkits: Optional[list[Toolkit]] = None, title: Optional[str] = None, description: Optional[str] = None):
        
        super().__init__(
            name=name,
            title=title,
            description=description,
            input_schema=merge(
                schema=prompt_schema(
                    description="use a prompt to generate content"),
                    additional_properties={
                    "output_schema" : { "type" : "object" }
                }
            ),
            output_schema=None)
            
        self._max_iterations = max_iterations

        self._planning_rules : list[str] = [
            *reasoning_rules
        ]

        self._responses = dict()

        self._llm_adapter = llm_adapter
        self._tool_adapter = tool_adapter 

        if toolkits == None:
            toolkits = []
        
        self.toolkits = toolkits
        

    async def init_chat_context(self):
        chat  = self._llm_adapter.create_chat_context()

        all_rules = self._planning_rules.copy()
        chat.append_rules(rules=all_rules)
        return chat


    async def ask(self, context: AgentCallContext, arguments: dict):

        dynamic_schema = arguments["output_schema"]

     
        class ResponseTool(Tool):
            def __init__(self, output_schema: dict, context: AgentCallContext, parent:PlanningResponder):

                output_schema = deepcopy(output_schema)
           
                schema = {
                    "type" : "object",
                    "additionalProperties" : False,
                    "required" : [ "summary", "data"], 
                    "properties" : {
                        "summary" :  {
                            "type" : "string",
                            "description" : "a summary of the data structure"
                        },
                        "data": output_schema
                    }
                }

                if "$defs" in output_schema:
                    schema["$defs"] = output_schema["$defs"]
                    del output_schema["$defs"]
                        
                super().__init__(
                    name="respond",
                    title="respond",
                    input_schema=schema,
                    description="send the response to the user",
                )
                self.parent = parent
                self.context = context 

            async def execute(self, *, context: ToolContext, **kwargs):
                self.parent._responses[self.context] = kwargs
                return TextResponse(text="the response was sent")

        class ResponseToolkit(Toolkit):
            def __init__(self, output_schema, context, parent):
                super().__init__(
                    name="meshagent.dynamic_response",
                    tools=[    
                        ResponseTool(output_schema=output_schema, context=context, parent=parent)
                    ]
                )

        context.toolkits.append(ResponseToolkit(output_schema=dynamic_schema, context=context, parent=self))
        
        execute = goto_next_step_message

        react = """
        based on what you know know, either execute the next task or formulate a new plan. If you have sufficient information to complete the task, return a final answer.
        """
    
        rs = reasoning_schema(description="uses tools", elements=[]).to_json()

        prompt = arguments["prompt"]

        context.chat.append_user_message(message=prompt)
        
        room = context.room
    
        i = 0
        while i < self._max_iterations:
            i += 1
            
            try:
                logger.info("COMPLETION STARTING: Step %s", i)
                
                toolkits = [
                    *self.toolkits,
                    *context.toolkits
                ]

                responses = await self._llm_adapter.next(context=context.chat, room=room, toolkits=toolkits, tool_adapter=self._tool_adapter, output_schema=rs)


                logger.info("COMPLETION RESPONSE %s", responses)
            
            except Exception as e:
                logger.error("Unable to execute reasoning completion task", exc_info=e)
                # retry
                raise(e)
            
            parsed = responses["response"]["data"][0]

            if "abort" in parsed:
                abort = parsed["abort"]
                reason = abort["reason"]
                raise AgentException(reason)
                
            elif "progress" in parsed:
                result = parsed["progress"]

                if "done" not in result or result["done"]:                   
                    if context not in self._responses:
                        context.chat.append_user_message("you must call the respond tool")
                        continue
                    
                    final_answer = self._responses.pop(context)
                    
                    logger.info("Done generating response %s", final_answer)
                    return final_answer["data"]
                
                else:
                    context.chat.append_user_message(execute)
                    continue
            elif "plan" in parsed:
                plan = parsed["plan"]
                context.chat.append_user_message(execute)
                continue
            else:
                logger.info("recieved invalid response, %s", parsed)
                context.chat.append_user_message("this response did not conform to the schema")
                continue

 




import logging
import asyncio
from typing import Optional

from meshagent.agents import TaskRunner
from meshagent.api.schema_document import Element,Text
from meshagent.api.room_server_client import RoomClient, MeshDocument
from meshagent.api.websocket_protocol import WebSocketClientProtocol

from typing import Optional

logger = logging.getLogger(__name__)


from meshagent.agents.agent import TaskRunner, AgentChatContext, AgentCallContext

class ListenerContext:
    def __init__(self, document: MeshDocument, room: RoomClient, call_context: AgentCallContext):
        self.document = document
        self.call_context = call_context
        self.room = room


# Notifies of new nodes or changed nodes in a document, the document must already exist
class Listener(TaskRunner):

    def __init__(self, *, name: str, wait_for_synchronize: bool = True, title: Optional[str] = None, description: Optional[str] = None):
        super().__init__(
            name=name,
            description=description,
            title=title,
            input_schema={
                "type" : "object",
                "required" : ["path"],
                "additionalProperties" : False,
                "properties" : {
                    "path" : {
                        "type" : "string",
                        "description" : "the path of the document to listen to"
                    }
                }
            },
            output_schema={
                "type" : "object",
                "additionalProperties" : False,
                "required" : [],
                "properties" : {
                }
            }
        )
        self.wait_for_synchronize = wait_for_synchronize
    
    async def on_listening_started(self, listener_context: ListenerContext):
        pass

    async def on_element_inserted(self, listener_context: ListenerContext, element: Element) -> bool:
        return False

    async def on_attribute_changed(self, listener_context: ListenerContext, element: Element, attribute: Optional[str]) -> bool:
        return False
    
    async def ask(self, *, context: AgentCallContext, arguments: dict):

        output_path = arguments["path"]
        room = context.room
        logger.info("Visitor connecting to %s", output_path)
        doc = await room.sync.open(path=output_path, create=True)
        try:
            listener_context = ListenerContext(
                document=doc,
                room=room,
                call_context=context,
            )
            logger.info("Visitor connected to %s", output_path)

            change_queue = list[Element | Text]()
            
            def append_children(node:Element):
                for child in node.get_children():
                    if child not in change_queue:
                        change_queue.append([child])
                    if isinstance(child, Element):
                        append_children(child)
            
            if self.wait_for_synchronize == False:
                change_queue.append([doc.root])
                append_children(doc.root)
            else:
                await doc.synchronized


            await self.on_listening_started(listener_context=listener_context)
                
            wait_for_changes = asyncio.Future()
            
            @doc.on("inserted")
            def on_inserted(e: Element):
                logger.info("element inserted %s", e.tag_name)
                if e not in change_queue:
                    change_queue.append([e])
                    append_children(e)
                
                if wait_for_changes.done() == False:
                    wait_for_changes.set_result(True)


            @doc.on("updated")
            def on_updated(e: Element, attribute: str):
                logger.info("element updated %s", e.tag_name)
                if e not in change_queue:                            
                    change_queue.append([e, attribute])
                    #
                    # append_children(e)
                
                if wait_for_changes.done() == False:
                    wait_for_changes.set_result(True)


            waiting_for_end = True
            while waiting_for_end:
                await wait_for_changes
            
                while len(change_queue) > 0:

                    change = change_queue.pop(0)
                    content = change[0]
                    if len(change) > 1:
                        done = await self.on_attribute_changed(listener_context, content, change[1])
                        if done:
                            waiting_for_end = False

                    else:
                        done = await self.on_element_inserted(listener_context, content)
                        if done:
                            waiting_for_end = False
                        
                    if content in change_queue:
                        change_queue.remove(content)

                    
                
                wait_for_changes = asyncio.Future()
        except Exception as e:
            logger.error("Failed to visit", exc_info=e)                
            raise

        finally:

            logger.info("vistor done")

            await asyncio.sleep(5)
            await room.sync.close(output_path)
            
            return {}



from meshagent.api import RoomClient, Requirement, RoomException, Participant
from meshagent.tools import validate_openai_schema, Toolkit
import json
from typing import Optional
import asyncio

async def generate_json(*, on_behalf_of: Optional[Participant] = None, room: RoomClient, prompt: Optional[str] = None, output_schema: dict, requires: Optional[list[Requirement]] = None) -> dict:
    
    # make sure agent is in the room before proceeding
    agent = None
    for i in range(10):
        agents = await room.agents.list_agents()
        for a in agents:
            if a.name == "meshagent.schema_planner":
                agent = a
                break
        
        if agent != None:
            break

        await asyncio.sleep(1)

    if agent == None:
        raise RoomException("unable to locate required agent (meshagent.schema_planner)")
        
    
    if prompt == None:
        prompt = f"ask me a series of questions to completely fill out the data structure described by this JSON schema ${json.dumps(output_schema)}. If you need to ask multiple questions, try to include all of them in a single form."

    validate_openai_schema(output_schema)
    return await room.agents.ask(
        on_behalf_of=on_behalf_of,
        agent="meshagent.schema_planner",
        requires = requires,
        arguments = {
            "prompt": prompt,
            "output_schema": output_schema,
        },
    )



from meshagent.agents import TaskRunner, RequiredToolkit, SingleRoomAgent
from meshagent.tools import Toolkit, Tool, ToolContext
from meshagent.api.room_server_client import TextDataType, VectorDataType, FloatDataType, IntDataType
from openai import AsyncOpenAI
from typing import Optional
from meshagent.api.chan import Chan

import hashlib
import chonkie
import asyncio
import logging

from functools import wraps

import os

# TODO: install chonkie, chonkie[semantic], openai

from concurrent.futures import ThreadPoolExecutor, as_completed

def _async_debounce(wait):
    def decorator(func):
        task = None

        @wraps(func)
        async def debounced(*args, **kwargs):
            nonlocal task

            async def call_func():
                await asyncio.sleep(wait)
                await func(*args, **kwargs)

            if task and not task.done():
                task.cancel()

            task = asyncio.create_task(call_func())
            return task

        return debounced

    return decorator

logger = logging.getLogger("indexer")


class Chunk:
    def __init__(self, *, text: str, start: int, end: int):
        self.text = text
        self.start = start
        self.end = end

class Chunker:

    async def chunk(self, *, text: str, max_length: Optional[int] = None) -> list[Chunk]:
        pass

class ChonkieChunker(Chunker):
    def __init__(self, chunker: Optional[chonkie.BaseChunker] = None):
        super().__init__()

        if chunker == None:
            chunker = chonkie.SemanticChunker()

        self._chunker = chunker

    async def chunk(self, *, text: str, max_length: Optional[int] = None) -> list[Chunk]:
        chunks = await asyncio.to_thread(self._chunker.chunk, text=text)
        mapped = []
        for chunk in chunks:
            mapped.append(Chunk(text=chunk.text, start=chunk.start_index, end=chunk.end_index))
        return mapped
    

class Embedder:
    def __init__(self, *, size: int, max_length: int):
        self.size = size
        self.max_length = max_length
    
    async def embed(self, *, text: str) -> list[float]:
        pass

class OpenAIEmbedder(Embedder):
    def __init__(self, *, size: int, max_length: int, model: str,  openai: Optional[AsyncOpenAI] = None, ):
        if openai == None:
            openai = AsyncOpenAI()
    
        self._openai = openai
        self._model = model

        super().__init__(size=size, max_length=max_length)

    
    async def embed(self, *, text):
        return (await self._openai.embeddings.create(input=text, model=self._model, encoding_format="float")).data[0].embedding
    


class RagTool(Tool):
    def __init__(self, *, name = "rag_search", table: str, title = "RAG search", description = "perform a RAG search", rules = None, thumbnail_url = None, embedder: Optional[Embedder] = None):
        
        self.table = table

        super().__init__(
            name=name,
            input_schema={
                "type":"object",
                "additionalProperties" : False,
                "required" : [
                    "query"
                ],
                "properties" : {
                    "query" : {
                        "type" : "string"
                    }
                }
            },
            title=title,
            description=description,
            rules=rules, thumbnail_url=thumbnail_url)
        
        self._embedder = embedder

    async def execute(self, context: ToolContext, query: str):
        
        if self._embedder == None:
            results = await context.room.database.search(
                table=self.table,
                text=query,
                limit=10
            )
        else:
            embedding = await self._embedder.embed(text=query)
            results = await context.room.database.search(
                table=self.table,
                text=query,
                vector=embedding,
                limit=10
            )

        results = list(map(lambda r: f"from {r["url"]}: {r["text"]}", results))

        return {
            "results" : results
        }
        

def open_ai_embedding_3_small():
    return OpenAIEmbedder(model="text-embedding-3-small", max_length=8191, size=1536)

def open_ai_embedding_3_large():
    return OpenAIEmbedder(model="text-embedding-3-large", max_length=8191, size=3072)

def open_ai_embedding_ada_2():
    return OpenAIEmbedder(model="text-embedding-ada-002", max_length=8191, size=1536)


class RagToolkit(Toolkit):
    def __init__(self, table: str, embedder:Optional[Embedder] = None):

        if embedder == None:
            embedder = open_ai_embedding_3_large()

        super().__init__(
            name="meshagent.rag",
            title="RAG",
            description="Searches against an index",
            tools=[
                RagTool(table=table, embedder=embedder)
            ]
        )


class FileIndexEvent:
    def __init__(self, *, path: str, deleted: bool):
        self.path = path
        self.deleted = deleted

class StorageIndexer(SingleRoomAgent):

    def __init__(
        self,
        *,
        name,
        title=None,
        description=None,
        requires=None,
        labels = None, 
        chunker: Optional[Chunker] = None,
        embedder:Optional[Embedder] = None,
        table: str = "storage_index",
      ):
        super().__init__(name=name, title=title, description=description, requires=requires, labels=labels)

        self._chan = Chan[FileIndexEvent]()
        
        if chunker == None:
            chunker = ChonkieChunker()

        if embedder == None:
            embedder = open_ai_embedding_3_large()

        self.chunker = chunker
        self.embedder = embedder
        self.table = table
        self._vector_index_created = False
        self._fts_created = False

    async def read_file(self, *, path: str) -> str | None:
        pass
    
    @_async_debounce(10)
    async def refresh_index(self):

        self.room.developer.log_nowait(type="indexer.rebuild", data={})

        indexes = await self.room.database.list_indexes(table=self.table)

        logger.info(f"existing indexes {indexes}")
        
        for index in indexes:

            if "embedding" in index["columns"]:
                self._vector_index_created = True
                
            if "text" in index["columns"]:
                self._fts_created  = True

        if self._vector_index_created == False:
            try:
                logger.info("attempting to create embedding index")
                await self.room.database.create_vector_index(table=self.table, column="embedding", replace=False)
                self._vector_index_created = True
            except Exception as e:
                # Will fail if there aren't enough rows
                pass

        if self._fts_created == False:
            try:
                logger.info("attempting to create fts index")
                await self.room.database.create_full_text_search_index(table=self.table, column="text", replace=False)
                self._fts_created = True
            except Exception as e:
                # Will fail if there aren't enough rows
                pass

        if self._fts_created == True or self._vector_index_created == True:
            logger.info("optimizing existing index")
            await self.room.database.optimize(table=self.table)
        

    async def start(self, *, room):
        await super().start(room=room)

        room.storage.on("file.updated", self._on_file_updated)
        room.storage.on("file.deleted", self._on_file_deleted)

        await room.database.create_table_with_schema(
            name=self.table,
            schema={
                "url" : TextDataType(),
                "text" : TextDataType(),
                "embedding" : VectorDataType(
                    size=self.embedder.size,
                    element_type=FloatDataType()
                ),
                "sha" : TextDataType(),
            },
            mode="create_if_not_exists",
            data=None
        )

      
        def index_task(task: asyncio.Task):

            try:
                result = task.result()
            except Exception as e:
                logger.error("Index task failed", exc_info=e)
            

        self._index_task = asyncio.create_task(self._indexer())
        self._index_task.add_done_callback(index_task)
    
    async def stop(self):
        await super().stop()
        await self._chan.close()
        
           
    async def _indexer(self):

        async for e in self._chan:

            try:
                if e.deleted:

                    # todo: consider using sql_alchemy or a library to do the escaping
                    def escape_sql_string(value):
                        if not isinstance(value, str):
                            raise TypeError("Input must be a string")
                        return value.replace("'", "''")

                    self.room.developer.log_nowait(type="indexer.delete", data={"path": e.path})
                    await self.room.database.delete(table=self.table, where=f"url='{escape_sql_string(e.path)}'")
                    

                else:

                    self.room.developer.log_nowait(type="indexer.index", data={"path": e.path})
                    

                    async def lookup_or_embed(*, sha: str, text: str) -> list[float]:

                        # if we already indexed this chunk, lets use the existing embedding instead of generating a new one
                        results = await self.room.database.search(
                            table=self.table,
                            where={
                                "sha" : sha,
                            },
                            limit=1
                        )

                        if len(results) != 0:
                            logger.info(f"chunk found from {e.path} {sha}, reusing embedding")
                            return results[0]["embedding"]
                            
                        logger.info(f"chunk not found from {e.path} {sha}, generating embedding")
                                
                        return await self.embedder.embed(text=text)

                    basename = os.path.basename(e.path)

                    chunk_sha = hashlib.sha256(basename.encode("utf-8")).hexdigest()

                    rows = []
                    # let's make the filename it's own chunk
                    rows.append(
                        {
                            "url" : e.path,
                            "text" : basename,
                            "sha" : chunk_sha,
                            "embedding" : await lookup_or_embed(sha=chunk_sha, text=basename)
                        }
                    )
                        
                    
                    text = await self.read_file(path=e.path)
                    if text != None:
                        
                        # the content will be transformed into additional chunks
                        for chunk in await self.chunker.chunk(text=text, max_length = self.embedder.max_length):
                            logger.info(f"processing chunk from {e.path}: {chunk.start}")
                            chunk_sha = hashlib.sha256(chunk.text.encode("utf-8")).hexdigest()
                            rows.append(
                                {
                                    "url" : e.path,
                                    "text" : chunk.text,
                                    "embedding" : await lookup_or_embed(sha=chunk_sha, text=chunk.text),
                                    "sha" : chunk_sha,
                                }
                            )
                    await self.room.database.merge(table=self.table, on="sha", records=rows)
                    await self.refresh_index()

                    
                
            except Exception as e:
                logger.error("error while indexing", exc_info=e)


    def _on_file_deleted(self, path: str, participant_id: str):
        self._chan.send_nowait(FileIndexEvent(path=path, deleted=True))
        
    def _on_file_updated(self, path: str, participant_id: str):
        self._chan.send_nowait(FileIndexEvent(path=path, deleted=False))

        
    


class SiteIndexer(TaskRunner):

    def __init__(self,
         *, 
        name,
        chunker: Optional[Chunker] = None,
        embedder:Optional[Embedder] = None,
        title=None,
        description=None,
        requires=None,
        supports_tools = None,
        labels: Optional[list[str]] = None
     
    ):
        
        if chunker == None:
            chunker = ChonkieChunker()

        if embedder == None:
            embedder = open_ai_embedding_3_large()

        self.chunker = chunker
        self.embedder = embedder

        super().__init__(
            name=name,
            title=title,
            description=description,
            requires=[
                RequiredToolkit(
                    name="meshagent.firecrawl",
                    tools=[
                        "firecrawl_queue"
                    ]
                ),
            ],
            supports_tools=supports_tools,
            input_schema={
                "type" : "object",
                "required" : [
                    "queue", "table", "url"
                ],
                "additionalProperties" : False,
                "properties" : {
                    "queue" : {
                        "type" : "string",
                        "description" : "default: firecrawl"
                    },
                    "table" : {
                        "type" : "string",
                        "description" : "default: index"
                    },
                    "url" : {
                        "type" : "string",
                        "description" : "default: index"
                    }
                }
            },
            output_schema={
                "type" : "object",
                "required" : [],
                "additionalProperties" : False,
                "properties" : {},
            },
            labels=labels
        )


    async def ask(self, *, context, arguments):
        
        queue = arguments["queue"]
        table = arguments["table"]
        url = arguments["url"]

        tables = await context.room.database.list_tables()

        exists = False
        try:
            exists = tables.index(table)
        except ValueError:
            pass


        async def lookup_or_embed(*, sha: str, text: str) -> list[float]:

            # if we already indexed this chunk, lets use the existing embedding instead of generating a new one
            if exists:

                results = await self.room.database.search(
                    table=self.table,
                    where={
                        "sha" : sha,
                    },
                    limit=1
                )


                if len(results) != 0:
                    logger.info(f"chunk found from {url} {sha}, reusing embedding")
                    return results[0]["embedding"]
                
            logger.info(f"chunk not found from {url} {sha}, generating embedding")
                    
            return await self.embedder.embed(text=text)
            
        
        async def crawl():
            logger.info(f"starting to crawl: {url}")
            await context.room.agents.invoke_tool(
                toolkit="meshagent.firecrawl",
                tool="firecrawl_queue",
                arguments={
                    "url" : url,
                    "queue": queue,
                    "limit" : 100
                })
            
            logger.info(f"done with crawl: {url}")
            await context.room.queues.send(name=queue, message={ "done" : True })
            
        def crawl_done(task: asyncio.Task):
            try:
                task.result()
            except Exception as e:
                logger.error("crawl failed", exc_info=e)


        crawl_task = asyncio.create_task(crawl())
        crawl_task.add_done_callback(crawl_done)
        
        rows = []

        id = 0
        
        while True:
            message = await context.room.queues.receive(name=queue, create=True, wait=True)
           
            if message == None:
                break

            if message.get("type", None) == "crawl.completed":
                break
            
            if "data" in message:
                for data in message["data"]:
                    try:
                        url : str  = data["metadata"]["url"]
                        text : str = data["markdown"]
                        title : str  = data["metadata"]["title"]
                        title_sha : str  = hashlib.sha256(text.encode("utf-8")).hexdigest()

                        logger.info(f"processing crawled page: {url}")
                        
                        # let's make the title it's own chunk
                        rows.append(
                                {
                                    "id" : id,
                                    "url" : url,
                                    "text" : title,
                                    "sha" : title_sha,
                                    "embedding" : await lookup_or_embed(sha=title_sha, text=title)
                                }
                            )
                            
                        id = id + 1
                        
                        # the content will be transformed into additional chunks
                        for chunk in await self.chunker.chunk(text=text, max_length = self.embedder.max_length):
                            logger.info(f"processing chunk from {url}: {chunk.text}")
                            chunk_sha = hashlib.sha256(chunk.text.encode("utf-8")).hexdigest()
                            rows.append(
                                {
                                    "id" : id,
                                    "url" : url,
                                    "text" : chunk.text,
                                    "embedding" : await lookup_or_embed(sha=chunk_sha, text=chunk.text)
                                }
                            )
                            
                            id = id + 1

                    except Exception as e:
                        logger.error(f"failed to process: {url}", exc_info=e)

        logger.info(f"saving crawl: {url}")
            
        await context.room.database.create_table_with_schema(
            name=table,
            schema={
                "id" : IntDataType(),
                "url" : TextDataType(),
                "text" : TextDataType(),
                "embedding" : VectorDataType(
                    size=self.embedder.size,
                    element_type=FloatDataType()
                ),
                "sha" : TextDataType(),
            },
            mode="overwrite",
            data=rows
        )

        if len(rows) > 255:
            await context.room.database.create_vector_index(table=table, column="embedding")

        await context.room.database.create_full_text_search_index(table=table, column="text")

        return {}



from meshagent.api.schema import MeshSchema, ElementType,  ChildProperty, ValueProperty

gallery_schema = MeshSchema(
    root_tag_name="gallery",
    elements=[      
        ElementType(
            tag_name="gallery", 
            description="a media gallery",
            properties=[
                ChildProperty(name="content", description="the content in the gallery", child_tag_names=[
                    "file"
                ]),
            ]
        ),   
        ElementType(
            tag_name="file", 
            description="an image",
            properties=[
                ValueProperty(name="path", description="reference a file that was output from one of the steps. use to display the content of a file or image.", type="string"),
                ValueProperty(name="description", description="a description of the image", type="string"),
            ]
        ),
    ]
)





from meshagent.api.schema import ValueProperty, ChildProperty
from meshagent.tools import Toolkit
from meshagent.agents.writer import Writer, WriterContext
from meshagent.agents.adapter import LLMAdapter, ToolResponseAdapter
from meshagent.api.schema_util import prompt_schema, merge
from typing import Optional
from meshagent.api import Requirement

import logging

logger = logging.getLogger("planning_agent")



class SingleShotWriter(Writer):

    def __init__(self,  name: str,  llm_adapter: LLMAdapter, tool_adapter:  Optional[ToolResponseAdapter] = None, description: Optional[str] = None, title: Optional[str] = None, rules:Optional[list[str]] = None, requires: Optional[list[Requirement]] = None, toolkits : Optional[list[Toolkit]] = None, supports_tools: Optional[bool] = None, labels: Optional[list[str]] = None, ):
        super().__init__(
            name=name,
            description=description,
            title=title,
            input_schema=merge(
                schema=prompt_schema(description="use a prompt to generate content"),
                additional_properties={
                    "path" : { "type" : "string" }
                }),
                output_schema={
                    "type" : "object",
                    "additionalProperties" : False,
                    "required" : [],
                    "properties" : {
                    }
                },
            requires=requires,
            supports_tools=supports_tools,
        )
        self._rules = rules
        self._llm_adapter = llm_adapter
        self._tool_adapter = tool_adapter 
        if toolkits == None:
            toolkits = []
        self._toolkits = toolkits
        self._labels = labels
    

    async def init_chat_context(self):
        context = self._llm_adapter.create_chat_context()
        context.append_rules(self._rules)
        return context

    async def write(self, writer_context: WriterContext, arguments: dict):

        arguments = arguments.copy()
        self.pop_path(arguments=arguments)

        prompt = arguments["prompt"]

        writer_context.call_context.chat.append_user_message(message=prompt)
            
        toolkits = [
            *self._toolkits,
            *writer_context.call_context.toolkits
        ]
        
        try:
            logger.info("SINGLE SHOT COMPLETION STARTING")
            
            response = await self._llm_adapter.next(context=writer_context.call_context.chat, room=writer_context.room, toolkits=toolkits, tool_adapter=self._tool_adapter, output_schema=writer_context.document.schema.to_json())

            logger.info("SINGLE SHOT COMPLETION RESPONSE %s", response)
        
        except Exception as e:
            logger.error("Unable to execute reasoning completion task", exc_info=e)
            # retry
            raise(e)
        
        document = writer_context.document
        response = response[document.schema.root.tag_name]

        for p in document.schema.root.properties:
            if isinstance(p, ValueProperty):
                document.root[p.name] = response[p.name]
            elif isinstance(p, ChildProperty):
                for value in response[p.name]:
                    document.root.append_json(value)
            else:
                raise Exception("Unexpected property type")

        return {}

 



from meshagent.api.schema import MeshSchema, ElementType,  ChildProperty, ValueProperty

document_schema = MeshSchema(
    root_tag_name="document",
    elements=[      
        ElementType(
            tag_name="document", 
            description="a document",
            properties=[
                ChildProperty(name="content", description="the content in the document", child_tag_names=[
                    "heading", "body", "file"
                ]),
            ]
        ),   
        ElementType(
            tag_name="heading", 
            description="heading",
            properties=[
                ValueProperty(name="text", description="the text of the heading", type="string"),
            ]
        ),          
        ElementType(
            tag_name="body", 
            description="text",
            properties=[                    
                ValueProperty(name="text", description="body content in markdown format", type="string"),
            ]
        ),
        ElementType(
            tag_name="file", 
            description="an image",
            properties=[
                ValueProperty(name="name", description="reference a file that was output from one of the steps. use to display the content of a file or image.", type="string"),
            ]
        ),
    ]
)




from meshagent.api.schema import ElementProperty, ValueProperty, MeshSchema, ChildProperty, ElementType
from meshagent.api.schema_document import Document

schema_schema = MeshSchema(
    root_tag_name="schema",
    elements=[
        ElementType(tag_name="schema", description="defines an xml document representing a json schema", properties=[
            ValueProperty(name="root_tag_name", description="the root tag name in this document (must match one of the elements defined in the schema)", type="string"),            
            ChildProperty(name="elements", description="the configuration of this sandbox", child_tag_names=[
                "element"
            ]),
        ]),
        
        ElementType(tag_name="element", description="an element inside the document", properties=[
            ValueProperty(name="tag_name", description="the tag name used by this element. Element names are case-sensitive. Element names cannot contain spaces. Element names must start with a letter or underscore. Element names cannot start with the letters xml (or XML, or Xml, etc) Element names can contain letters, digits, hyphens, underscores, and periods.", type="string"),            
            ChildProperty(name="schema", description="the attributes of this node", child_tag_names=[
                "attribute", "child_elements"
            ]),
            ValueProperty(name="description", type="string"),
        ]),
        
        ElementType(tag_name="attribute", properties=[
            ValueProperty(name="name", description="the attribute name. Attribute names are case-sensitive. Attribute names cannot contain spaces. Attribute names must start with a letter or underscore. Attribute names cannot start with the letters xml (or XML, or Xml, etc) Attribute names can contain letters, digits, hyphens, underscores, and periods.", type="string"),
            ValueProperty(name="type", description="the type of value stored by this attribute", type="string", enum= [ "string", "boolean", "number" ]),
            ValueProperty(name="description", type="string"),
        ]),
        
        ElementType(tag_name="child_elements", properties=[
            ValueProperty(name="name", description="the name of the property to store the children in. Propterty names cannot contain spaces. Propterty names must start with a letter or underscore. Propterty names cannot start with the letters xml (or XML, or Xml, etc) Propterty names can contain letters, digits, hyphens, and underscores.", type="string"),
            ValueProperty(name="description", type="string"),
            ChildProperty(name="children", description="the tag name of an element. The element must have a corresponding element definition under the schema elements", child_tag_names=["element_reference"]),            
        ]),

        ElementType(tag_name="element_reference", description="an allowed child element for an element, should only used when a schema can contain multiple items of a specific type", properties=[
            ValueProperty(name="tag_name", description="the tag name of an element. The element must have a corresponding element definition under the schema elements", type="string"),            
        ]),
    ],
)

def schema_from_schema_document(document: Document):
    root_tag_name = document.root["root_tag_name"]

    elements = list[ElementType]()

    for element in document.root.get_children():

        properties = list[ElementProperty]()
    
        for prop in element.get_children():
            name = prop["name"]
            description = prop["description"]
            type = prop["type"]
            
            if prop.tag_name == "attribute":
                properties.append(ValueProperty(
                    name = name,
                    type = type,
                    description = description
                ))
            elif prop.tag_name == "child_elements":
                child_tag_names = []
                
                for child in prop.get_children():
                    child_tag_names.append(child["tag_name"])

                properties.append(ChildProperty(
                    name=name,
                    child_tag_names = child_tag_names
                ))

        et = ElementType(tag_name = element["tag_name"], properties = properties, description = element["description"])
        elements.append(et)

    doc = MeshSchema(root_tag_name=root_tag_name, elements=elements)
    return doc



import json
from copy import deepcopy

from typing import Optional

from meshagent.api.messaging import unpack_message, pack_message
from meshagent.api.room_server_client import RoomException, RequiredToolkit, Requirement, RequiredSchema
from meshagent.api import WebSocketClientProtocol, ToolDescription, ToolkitDescription, Participant, RemoteParticipant, meshagent_base_url, StorageEntry, JsonResponse
from meshagent.api.protocol import Protocol
from meshagent.tools.toolkit import Toolkit, Tool, ToolContext, toolkit_factory, register_toolkit_factory
from meshagent.api.room_server_client import RoomClient, RoomException
from jsonschema import validate
from .context import AgentCallContext, AgentChatContext
from meshagent.api.schema_util import no_arguments_schema
from meshagent.api import MeshSchema
import logging
import asyncio
from typing import Optional, Dict, Callable, Coroutine

logger = logging.getLogger("agent")


class AgentException(RoomException):
    pass


class RoomTool(Tool):
    def __init__(self, *, toolkit_name: str, name, input_schema, title = None, description = None, rules = None, thumbnail_url = None, participant_id: Optional[str] = None, on_behalf_of_id: Optional[str] = None, defs: Optional[dict] = None):
        self._toolkit_name = toolkit_name
        self._participant_id = participant_id
        self._on_behalf_of_id = on_behalf_of_id

        super().__init__(name=name, input_schema=input_schema, title=title, description=description, rules=rules, thumbnail_url=thumbnail_url, defs=defs)

    async def execute(self, context, **kwargs):
       return await context.room.agents.invoke_tool(
            toolkit=self._toolkit_name,
            tool=self.name,
            participant_id=self._participant_id,
            on_behalf_of_id=self._on_behalf_of_id,
            arguments=kwargs,
    
        )
    
class Agent:

    def __init__(self, *, name: str, title: Optional[str] = None, description: Optional[str] = None, requires: Optional[list[Requirement]] = None, labels: Optional[list[str]] = None):
       
        self._name = name
        if title == None:
            title = name
        self._title = title
        if description == None:
            description = ""

        self._description = description
        if requires == None:
            requires = []

        self.init_requirements(requires)
        self._requires = requires
        
        if labels == None:
            labels = []
        
        self._labels = labels

    @property
    def name(self):
        return self._name
    
    @property
    def description(self):
        return self._description
    
    @property
    def title(self):
        return self._title
    
    @property
    def requires(self):
        return self._requires
    
    @property
    def labels(self):
        return self._labels
    
    def init_requirements(self, requires): ...
    
    async def init_chat_context(self) -> AgentChatContext:
        return AgentChatContext() 

    def to_json(self) -> dict:
        return {
            "name" : self.name,
            "title" : self.title,
            "description" : self.description,
            "requires" : list(map(lambda x: x.to_json(), self.requires)),
            "labels" : self.labels
        }

class SingleRoomAgent(Agent):

    def __init__(self, *, name, title = None, description = None, requires = None, labels: Optional[list[str]] = None):
        super().__init__(name=name, title=title, description=description, requires=requires, labels=labels)
        self._room = None


    async def start(self, *, room: RoomClient) -> None:

        if self._room != None:
            raise RoomException("room is already started")
        
        self._room = room

        await self.install_requirements()


    async def stop(self) -> None:
        self._room = None
        pass

    @property
    def room(self):
        return self._room
    
    async def install_requirements(self, participant_id: Optional[str] = None):
       
        schemas_by_name = dict[str, StorageEntry]()

        schemas = await self._room.storage.list(path=".schemas")
        
        for schema in schemas:
            schemas_by_name[schema.name] = schema

        toolkits_by_name = dict[str, ToolkitDescription]()

        visible_tools = await self._room.agents.list_toolkits(participant_id=participant_id)
             
        for toolkit_description in visible_tools:
            toolkits_by_name[toolkit_description.name] = toolkit_description

        installed = False

        for requirement in self.requires:
            
            if isinstance(requirement, RequiredToolkit):
                
                if toolkit_factory(requirement.name) != None:
                    # no need to install something we can create from a factory
                    continue

                if requirement.name == "ui":
                    # TODO: maybe requirements can be marked as non installable?
                    continue
                
                if requirement.name not in toolkits_by_name:
                    
                    installed = True

                    logger.info(f"Installing required tool {requirement.name}")
                    
                    if requirement.name.startswith("https://"):
                        url = requirement.name
                    else:
                        url = f"{meshagent_base_url()}/toolkits/{requirement.name}"

                    await self._room.agents.make_call(url=url, name=requirement.name, arguments={})
            
            elif isinstance(requirement, RequiredSchema):

                if requirement.name not in schemas_by_name:

                    installed = True

                    schema_path = f".schemas/{requirement.name}.json"
                    
                    if await self._room.storage.exists(path=schema_path):
                        # Schema is already in the room
                        pass
                    else:
                        logger.info(f"Installing required tool {requirement.name}")
                        
                        if requirement.name.startswith("https://"):
                            url = requirement.name
                        else:
                            url = f"{meshagent_base_url()}/schemas/{requirement.name}"
                    
                        await self._room.agents.make_call(url=url, name=requirement.name, arguments={})

            else:
                raise RoomException("unsupported requirement")

        if installed:
            await asyncio.sleep(5)

    async def get_required_toolkits(self, context: ToolContext) -> list[Toolkit]:


        tool_target = context.caller
        if context.on_behalf_of != None:
            tool_target = context.on_behalf_of

        toolkits_by_name = dict[str, ToolkitDescription]()

        toolkits = list[Toolkit]()

        visible_tools = await self._room.agents.list_toolkits(participant_id=tool_target.id)

        for toolkit_description in visible_tools:
            toolkits_by_name[toolkit_description.name] = toolkit_description


        for required_toolkit in self.requires:
            
            if isinstance(required_toolkit, RequiredToolkit):
                
                if toolkit_factory(required_toolkit.name) != None:
                
                    
                    toolkit = await toolkit_factory(required_toolkit.name)(context, required_toolkit)
                    toolkits.append(toolkit)
                    continue


                toolkit = toolkits_by_name.get(required_toolkit.name, None)
                if toolkit == None:
                    raise RoomException(f"unable to locate required toolkit {required_toolkit.name}")

                room_tools = list[RoomTool]()
            

                if required_toolkit.tools == None:

                    for tool_description in toolkit.tools:
                        tool = RoomTool(
                            on_behalf_of_id=tool_target.id,
                            toolkit_name=toolkit.name,
                            name=tool_description.name,
                            description=tool_description.description,
                            input_schema=tool_description.input_schema,
                            title=tool_description.title,
                            thumbnail_url=tool_description.thumbnail_url,
                            participant_id=tool_target.id,
                            defs = tool_description.defs
                        )
                        room_tools.append(tool)

                else:
                    tools_by_name = dict[str, ToolDescription]()
                    for tool_description in toolkit.tools:
                        tools_by_name[tool_description.name] = tool_description

                    for required_tool in required_toolkit.tools:
                        
                        tool_description = tools_by_name.get(required_tool, None)
                        if tool_description == None:
                            raise RoomException(f"unable to locate required tool {required_tool}")

                        tool = RoomTool(
                            on_behalf_of_id=tool_target.id,
                            toolkit_name=toolkit.name,
                            name=tool_description.name,
                            description=tool_description.description,
                            input_schema=tool_description.input_schema,
                            title=tool_description.title,
                            thumbnail_url=tool_description.thumbnail_url,
                            participant_id=tool_target.id,
                            defs = tool_description.defs
                        )
                        room_tools.append(tool)

                toolkits.append(Toolkit(
                    name = toolkit.name,
                    title = toolkit.title,
                    description = toolkit.description,
                    thumbnail_url = toolkit.thumbnail_url,
                    tools = room_tools,
                ))

        return toolkits
        
class TaskRunner(SingleRoomAgent):

    def __init__(self, *, name, title = None, description = None, requires = None, supports_tools : Optional[bool] = None, input_schema: dict, output_schema: dict, labels: Optional[list[str]] = None, toolkits: Optional[list[Toolkit]] = None):
        super().__init__(name=name, title=title, description=description, requires=requires, labels=labels)

        if toolkits == None:
            toolkits = []

        self._toolkits = toolkits

        self._registration_id = None     

        if input_schema == None:
            input_schema = no_arguments_schema(
                description="execute the agent",
            )

        if supports_tools == None:
            supports_tools = False
            
        self._supports_tools = supports_tools
        self._input_schema = input_schema
        self._output_schema = output_schema
        

        
    async def validate_arguments(self, arguments: dict):
        validate(arguments, self.input_schema)

    async def validate_response(self, response: dict):
        if self.output_schema != None:
            validate(response, self.output_schema)

    async def ask(self, *, context: AgentCallContext, arguments: dict) -> dict:
        raise Exception("Not implemented")
    
    @property
    def supports_tools(self):
        return self._supports_tools
        
    @property
    def input_schema(self):
        return self._input_schema

    @property
    def output_schema(self):
        return self._output_schema

    def to_json(self) -> dict:
        return {
            "name" : self.name,
            "title" : self.title,
            "description" : self.description,
            "input_schema" : self.input_schema,
            "output_schema" : self.output_schema,
            "requires" : list(map(lambda x: x.to_json(), self.requires)),
            "supports_tools" : self.supports_tools,
            "labels" : self.labels
        }
    
    async def _register(self):
      self._registration_id = (await self._room.send_request("agent.register_agent", {
        "name": self.name,
        "title" : self.title,
        "description" : self.description,
        "input_schema" : self.input_schema,
        "output_schema" : self.output_schema,
        "requires" : list(map(lambda x : x.to_json(), self.requires)),
        "supports_tools" : self.supports_tools,
        "labels" : self.labels
      }))["id"]
    

    async def _unregister(self):
      await self._room.send_request("agent.unregister_agent", {
        "id": self._registration_id
      })
      self._registration_id = None


    async def start(self, *, room: RoomClient):

        await super().start(room=room)

        self._room.protocol.register_handler("agent.ask", self._ask)
        await self._register()
            

    async def stop(self): 
        if self.room.protocol.is_open:
            await self._unregister()
        else:
            logger.info(f"disconnected '{self.name}' from room, this will automatically happen when all the users leave the room. agents will not keep the room open")

        self._room.protocol.unregister_handler("agent.ask", self._ask)

        await super().stop()

    async def _ask(self, protocol:Protocol, message_id:int, msg_type:str, data:bytes):
        
        
        async def worker():
            # Decode and parse the message
            message, _ = unpack_message(data)
            logger.info("agent got message %s", message)
            args = message["arguments"]
            task_id = message["task_id"] 
            context_json = message["context"]
            toolkits_json = message["toolkits"]
            
            #context_json = message["context"]

            chat_context = None

            try:
                chat_context = await self.init_chat_context()

                caller : Participant | None = None
                on_behalf_of : Participant | None = None
                on_behalf_of_id = message.get("on_behalf_of_id", None) 
    
                for participant in self._room.messaging.get_participants():
                    if message["caller_id"] == participant.id:
                        caller = participant
                        break
                
                    if on_behalf_of_id == participant.id:
                        on_behalf_of = participant
                        break

                if caller == None:
                    caller = RemoteParticipant(
                        id=message["caller_id"],
                        role="user",
                        attributes={}
                    )

                if on_behalf_of_id != None and on_behalf_of == None:
                    on_behalf_of = RemoteParticipant(
                        id=message["on_behalf_of_id"],
                        role="user",
                        attributes={}
                    )
               
                tool_target = caller
                if on_behalf_of != None:
                    tool_target = on_behalf_of

                
                toolkits = [
                    *self._toolkits,
                    *await self.get_required_toolkits(context=ToolContext(room=self.room, caller=caller, on_behalf_of=on_behalf_of, caller_context={ "chat": chat_context.to_json() }))
                ]

                context = AgentCallContext(chat=chat_context, room=self.room, caller=caller, on_behalf_of=on_behalf_of, toolkits=toolkits)
                 
                for toolkit_json in toolkits_json:
                    tools = []
                    for tool_json in toolkit_json["tools"]:
                        tools.append(RoomTool(
                            on_behalf_of_id=on_behalf_of_id,
                            participant_id=tool_target.id,
                            toolkit_name=toolkit_json["name"],
                            name=tool_json["name"],
                            title=tool_json["title"],
                            description=tool_json["description"],
                            input_schema=tool_json["input_schema"],
                            thumbnail_url=toolkit_json["thumbnail_url"],
                            defs=tool_json.get("defs", None)
                        ))

                    context.toolkits.append(Toolkit(
                        name=toolkit_json["name"],
                        title=toolkit_json["title"],
                        description=toolkit_json["description"],
                        thumbnail_url=toolkit_json["thumbnail_url"],
                        tools=tools,
                    ))


          
                response = await self.ask(context=context, arguments=args)
                
                await protocol.send(type="agent.ask_response", data=pack_message({ 
                    "task_id" : task_id,
                    "answer" : response,
                    "caller_context" : {
                        "chat" : chat_context.to_json()
                    }
                }))

            except Exception as e:
                logger.error("Task runner failed to complete task", exc_info=e)
                if chat_context != None:
                        
                    await protocol.send(type="agent.ask_response", data=pack_message({
                        "task_id" : task_id,
                        "error" : str(e),
                        "caller_context" : {
                            "chat" : chat_context.to_json()
                        }
                    }))
                else:
                    await protocol.send(type="agent.ask_response", data=pack_message({
                        "task_id" : task_id,
                        "error" : str(e),
                    }))

        def on_done(task: asyncio.Task):
            task.result()

        task = asyncio.create_task(worker())
        task.add_done_callback(on_done)    


class RunTaskTool(Tool):
    def __init__(self, *, name: str, agent_name: str, input_schema: dict, rules = None, thumbnail_url = None, title: Optional[str] = None, description: Optional[str]  = None):
        super().__init__(name=name, input_schema=input_schema, rules=rules, thumbnail_url=thumbnail_url, title=title, description=description)

        self._agent_name = agent_name

    async def execute(self, context: ToolContext, **kwargs):
        return await context.room.agents.ask(agent=self._agent_name, arguments=kwargs)
    

async def make_run_task_tool(context: ToolContext, toolkit: RequiredToolkit):

    agents = await context.room.agents.list_agents()
    tools = []
    for agent_name in toolkit.tools:
        agent = next((x for x in agents if x.name == agent_name), None)

        if agent == None:
            raise RoomException(f"agent was not found in the room {agent_name}")
        
        tools.append(
            RunTaskTool(
                agent_name=agent_name,
                name=f"run_{agent_name}",
                input_schema=agent.input_schema,
                title=agent.title,
                description=agent.description
            ))

    return Toolkit(
        name="agents",
        tools=tools
    )

register_toolkit_factory("agents", make_run_task_tool)



from meshagent.agents.agent import  AgentCallContext
from meshagent.api.room_server_client import RoomClient
from meshagent.api.schema_document import Document
from meshagent.api import RoomClient, RequiredToolkit, RequiredSchema, Requirement
#from meshagent.tools.document_tools import DocumentToolkit
from typing import Optional
import asyncio
import logging
from meshagent.agents.hosting import TaskRunner

logger = logging.getLogger("writer_agent")


class WriterContext:
    def __init__(self, room: RoomClient, document: Document, call_context: AgentCallContext, path: str):
        self._room = room
        self._document = document
        self._call_context = call_context 
        self._path = path

    @property
    def room(self) -> RoomClient:
        return self._room
    
    @property
    def path(self) -> str:
        return self._path
    
    @property
    def document(self) -> Document:
        return self._document
    
    @property
    def call_context(self) -> AgentCallContext:
        return self._call_context
    

# writes to a path, the document must already exist, optionally create the document if it does not exist
class Writer(TaskRunner):
    def __init__(self, name: str, input_schema:dict, output_schema:dict, create: bool = True, title: Optional[str] = None, description: Optional[str] = None, requires: Optional[list[Requirement]] = None, supports_tools: Optional[bool] = None):
        super().__init__(
            name=name,
            input_schema=input_schema,
            output_schema=output_schema,
            description=description,
            title=title,
            requires=requires,
            supports_tools=supports_tools
        )
        
        self._create = create

        if "path" not in input_schema["properties"]:
            raise Exception("input schema must include a 'path' property")
        
    def pop_path(self, arguments: dict):
        return arguments["path"]
    
    async def write(self, writer_context: WriterContext, arguments: dict) -> dict:
        pass
 
    async def ask(self, *, context: AgentCallContext, arguments: dict):
    
        path = arguments["path"]
        
        client = context.room
        document = await client.sync.open(path=path, create=self._create)

        writer_context = WriterContext(
            room=client,
            document=document,
            call_context=context,
            path=path
        )
    
        try:
            return await self.write(writer_context, arguments)
        finally:
            # TODO: need a way to wait for changes to be synchronized
            await asyncio.sleep(1)
            await client.sync.close(path=path)